@@ -10,15 +10,17 @@ use reqwest::{
 
 use atuin_common::{
     api::{
-        AddHistoryRequest, CountResponse, DeleteHistoryRequest, ErrorResponse, IndexResponse,
-        LoginRequest, LoginResponse, RegisterResponse, StatusResponse, SyncHistoryResponse,
+        AddHistoryRequest, ChangePasswordRequest, CountResponse, DeleteHistoryRequest,
+        ErrorResponse, LoginRequest, LoginResponse, RegisterResponse, StatusResponse,
+        SyncHistoryResponse,
     },
-    record::RecordIndex,
+    record::RecordStatus,
 };
 use atuin_common::{
     api::{ATUIN_CARGO_VERSION, ATUIN_HEADER_VERSION, ATUIN_VERSION},
-    record::{EncryptedData, HostId, Record, RecordId},
+    record::{EncryptedData, HostId, Record, RecordIdx},
 };
+
 use semver::Version;
 use time::format_description::well_known::Rfc3339;
 use time::OffsetDateTime;
@@ -97,7 +99,10 @@ pub async fn login(address: &str, req: LoginRequest) -> Result<LoginResponse> {
     Ok(session)
 }
 
+#[cfg(feature = "check-update")]
 pub async fn latest_version() -> Result<Version> {
+    use atuin_common::api::IndexResponse;
+
     let url = "https://api.atuin.sh";
     let client = reqwest::Client::new();
 
@@ -143,6 +148,28 @@ pub fn ensure_version(response: &Response) -> Result<bool> {
     Ok(true)
 }
 
+async fn handle_resp_error(resp: Response) -> Result<Response> {
+    let status = resp.status();
+
+    if status == StatusCode::SERVICE_UNAVAILABLE {
+        bail!(
+            "Service unavailable: check https://status.atuin.sh (or get in touch with your host)"
+        );
+    }
+
+    if status.is_client_error() {
+        let error = resp.json::<ErrorResponse>().await?.reason;
+        bail!("Could not fetch history, client error: {error}.")
+    } else if status.is_server_error() {
+        let error = resp.json::<ErrorResponse>().await?.reason;
+        bail!("There was an error with the atuin sync service: {error}.\nIf the problem persists, contact the host")
+    } else if !status.is_success() {
+        bail!("There was an error with the atuin sync service: Status {status:?}.\nIf the problem persists, contact the host")
+    }
+
+    Ok(resp)
+}
+
 impl<'a> Client<'a> {
     pub fn new(
         sync_addr: &'a str,
@@ -172,6 +199,7 @@ impl<'a> Client<'a> {
         let url = Url::parse(url.as_str())?;
 
         let resp = self.client.get(url).send().await?;
+        let resp = handle_resp_error(resp).await?;
 
         if !ensure_version(&resp)? {
             bail!("could not sync due to version mismatch");
@@ -191,15 +219,12 @@ impl<'a> Client<'a> {
         let url = Url::parse(url.as_str())?;
 
         let resp = self.client.get(url).send().await?;
+        let resp = handle_resp_error(resp).await?;
 
         if !ensure_version(&resp)? {
             bail!("could not sync due to version mismatch");
         }
 
-        if resp.status() != StatusCode::OK {
-            bail!("failed to get status (are you logged in?)");
-        }
-
         let status = resp.json::<StatusResponse>().await?;
 
         Ok(status)
@@ -228,27 +253,18 @@ impl<'a> Client<'a> {
         );
 
         let resp = self.client.get(url).send().await?;
+        let resp = handle_resp_error(resp).await?;
 
-        let status = resp.status();
-        if status.is_success() {
-            let history = resp.json::<SyncHistoryResponse>().await?;
-            Ok(history)
-        } else if status.is_client_error() {
-            let error = resp.json::<ErrorResponse>().await?.reason;
-            bail!("Could not fetch history: {error}.")
-        } else if status.is_server_error() {
-            let error = resp.json::<ErrorResponse>().await?.reason;
-            bail!("There was an error with the atuin sync service: {error}.\nIf the problem persists, contact the host")
-        } else {
-            bail!("There was an error with the atuin sync service: Status {status:?}.\nIf the problem persists, contact the host")
-        }
+        let history = resp.json::<SyncHistoryResponse>().await?;
+        Ok(history)
     }
 
     pub async fn post_history(&self, history: &[AddHistoryRequest]) -> Result<()> {
         let url = format!("{}/history", self.sync_addr);
         let url = Url::parse(url.as_str())?;
 
-        self.client.post(url).json(history).send().await?;
+        let resp = self.client.post(url).json(history).send().await?;
+        handle_resp_error(resp).await?;
 
         Ok(())
     }
@@ -257,20 +273,39 @@ impl<'a> Client<'a> {
         let url = format!("{}/history", self.sync_addr);
         let url = Url::parse(url.as_str())?;
 
-        self.client
+        let resp = self
+            .client
             .delete(url)
-            .json(&DeleteHistoryRequest { client_id: h.id })
+            .json(&DeleteHistoryRequest {
+                client_id: h.id.to_string(),
+            })
             .send()
             .await?;
 
+        handle_resp_error(resp).await?;
+
+        Ok(())
+    }
+
+    pub async fn delete_store(&self) -> Result<()> {
+        let url = format!("{}/api/v0/store", self.sync_addr);
+        let url = Url::parse(url.as_str())?;
+
+        let resp = self.client.delete(url).send().await?;
+
+        handle_resp_error(resp).await?;
+
         Ok(())
     }
 
     pub async fn post_records(&self, records: &[Record<EncryptedData>]) -> Result<()> {
-        let url = format!("{}/record", self.sync_addr);
+        let url = format!("{}/api/v0/record", self.sync_addr);
         let url = Url::parse(url.as_str())?;
 
-        self.client.post(url).json(records).send().await?;
+        debug!("uploading {} records to {url}", records.len());
+
+        let resp = self.client.post(url).json(records).send().await?;
+        handle_resp_error(resp).await?;
 
         Ok(())
     }
@@ -279,37 +314,37 @@ impl<'a> Client<'a> {
         &self,
         host: HostId,
         tag: String,
-        start: Option<RecordId>,
+        start: RecordIdx,
         count: u64,
     ) -> Result<Vec<Record<EncryptedData>>> {
-        let url = format!(
-            "{}/record/next?host={}&tag={}&count={}",
-            self.sync_addr, host.0, tag, count
+        debug!(
+            "fetching record/s from host {}/{}/{}",
+            host.0.to_string(),
+            tag,
+            start
         );
-        let mut url = Url::parse(url.as_str())?;
 
-        if let Some(start) = start {
-            url.set_query(Some(
-                format!(
-                    "host={}&tag={}&count={}&start={}",
-                    host.0, tag, count, start.0
-                )
-                .as_str(),
-            ));
-        }
+        let url = format!(
+            "{}/api/v0/record/next?host={}&tag={}&count={}&start={}",
+            self.sync_addr, host.0, tag, count, start
+        );
+
+        let url = Url::parse(url.as_str())?;
 
         let resp = self.client.get(url).send().await?;
+        let resp = handle_resp_error(resp).await?;
 
         let records = resp.json::<Vec<Record<EncryptedData>>>().await?;
 
         Ok(records)
     }
 
-    pub async fn record_index(&self) -> Result<RecordIndex> {
-        let url = format!("{}/record", self.sync_addr);
+    pub async fn record_status(&self) -> Result<RecordStatus> {
+        let url = format!("{}/api/v0/record", self.sync_addr);
         let url = Url::parse(url.as_str())?;
 
         let resp = self.client.get(url).send().await?;
+        let resp = handle_resp_error(resp).await?;
 
         if !ensure_version(&resp)? {
             bail!("could not sync records due to version mismatch");
@@ -317,6 +352,8 @@ impl<'a> Client<'a> {
 
         let index = resp.json().await?;
 
+        debug!("got remote index {:?}", index);
+
         Ok(index)
     }
 
@@ -334,4 +371,35 @@ impl<'a> Client<'a> {
             bail!("Unknown error");
         }
     }
+
+    pub async fn change_password(
+        &self,
+        current_password: String,
+        new_password: String,
+    ) -> Result<()> {
+        let url = format!("{}/account/password", self.sync_addr);
+        let url = Url::parse(url.as_str())?;
+
+        let resp = self
+            .client
+            .patch(url)
+            .json(&ChangePasswordRequest {
+                current_password,
+                new_password,
+            })
+            .send()
+            .await?;
+
+        dbg!(&resp);
+
+        if resp.status() == 401 {
+            bail!("current password is incorrect")
+        } else if resp.status() == 403 {
+            bail!("invalid login details");
+        } else if resp.status() == 200 {
+            Ok(())
+        } else {
+            bail!("Unknown error");
+        }
+    }
 }
